﻿using UnityEngine;
using Windows.Kinect;
using System.Runtime.InteropServices;
using System;
using System.Collections;
using System.Linq;
using System.Collections.Generic;

public class CoordinateMapperManager : MonoBehaviour
{
	private DepthSpacePoint[] _DepthCoordinates;

    private KinectModule.SourceBuffer _SourceBuffer;
    private KinectModule.KinectAccessManager _KinectAccessManager;

	long frameCount = 0;

	double elapsedCounter = 0.0;
	double fps = 0.0;
	
	Texture2D _ColorRGBX;

    public Texture2D ColorTexture
    { get { return _ColorRGBX; } }
    public byte[] BodyIndexBuffer
    { get { return _SourceBuffer.BodyIndexBuffer; } }
    public DepthSpacePoint[] DepthCoordinates
    { get { return _DepthCoordinates; } }
    public Body[] BodyBuffer
    { get { return _SourceBuffer.BodyBuffer; } }

    void Awake()
	{
        _ColorRGBX = new Texture2D (KinectModule.KinectConstants.ColorWidth,
                                    KinectModule.KinectConstants.ColorHeight,
                                    TextureFormat.RGBA32, false);

		_DepthCoordinates = new DepthSpacePoint[KinectModule.KinectConstants.ColorWidth *
                                                KinectModule.KinectConstants.ColorHeight];

        _KinectAccessManager.Access();
    }

	void OnGUI () 
	{
		GUI.Box (new Rect(10, 10, 200, 30), "FPS: " + fps.ToString("0.00"));

		if (_KinectAccessManager.NullFrame)
			GUI.Box (new Rect(10, 50, 200, 30), "NULL MSFR Frame");
	}

    void MapColorToDepth()
    {
        var DepthData = GCHandle.Alloc(_SourceBuffer.DepthBuffer, GCHandleType.Pinned);
        var DepthCoordinatesData = GCHandle.Alloc(_DepthCoordinates, GCHandleType.Pinned);

        _KinectAccessManager.CoordinateMapper.MapColorFrameToDepthSpaceUsingIntPtr(
            DepthData.AddrOfPinnedObject(),
            (uint)_SourceBuffer.DepthBuffer.Length * sizeof(ushort),
            DepthCoordinatesData.AddrOfPinnedObject(),
            (uint)_DepthCoordinates.Length);

        DepthCoordinatesData.Free();
        DepthData.Free();
    }

    void UpdateTexture()
    {
        _ColorRGBX.LoadRawTextureData(_SourceBuffer.ColorBuffer);
        _ColorRGBX.Apply();
    }

    void GetFPS()
    {
        elapsedCounter += Time.deltaTime;
        if (elapsedCounter > 1.0)
        {
            fps = frameCount / elapsedCounter;
            frameCount = 0;
            elapsedCounter = 0.0;
        }
        frameCount++;
    }
	
	void Update()
	{
        GetFPS();

		var sourceFrame = new KinectModule.RealMultiSourceFrame(_KinectAccessManager.SourceFrame); ;
        
        if (!_KinectAccessManager.NullFrame)
        {
            _SourceBuffer.UpdateBuffers(sourceFrame);

            MapColorToDepth();

            UpdateTexture();
        }
	}

	void OnApplicationQuit()
	{
        _KinectAccessManager.Dispose();
	}
}

namespace KinectModule
{
    public class KinectConstants
    {
        public const int DepthWidth = 512;
        public const int DepthHeight = 424;
        public const int ColorWidth = 1920;
        public const int ColorHeight = 1080;
        public const int BodyCount = 6;
        public const int ColorCount = 4;
    }

    public class SourceBuffer
    {
        private byte[] _ColorBuffer =     new byte[KinectConstants.ColorWidth *
                                                   KinectConstants.ColorHeight *
                                                   KinectConstants.ColorCount],
                       _BodyIndexBuffer = new byte[KinectConstants.DepthWidth *
                                                   KinectConstants.DepthHeight];
        private ushort[] _DepthBuffer = new ushort[KinectConstants.DepthWidth *
                                                   KinectConstants.DepthHeight];
        private Body[] _BodyBuffer =      new Body[KinectConstants.BodyCount];

        // 1. Do not return null values.
        // 2. The updated value should reflect the data of the source frame.

        public byte[] ColorBuffer
        { get { return _ColorBuffer; } }
        public byte[] BodyIndexBuffer
        { get { return _BodyIndexBuffer; } }
        public ushort[] DepthBuffer
        { get { return _DepthBuffer; } }
        public Body[] BodyBuffer
        { get { return _BodyBuffer; } }

        // 1. The data of the source frame must be reflected in the buffer.

        public void UpdateBuffers(IMultiSourceFrame sourceFrame)
        {
            sourceFrame.LastColorFrame.CopyData(_ColorBuffer);
            sourceFrame.LastBodyIndexFrame.CopyData(_BodyIndexBuffer);
            sourceFrame.LastDepthFrame.CopyData(_DepthBuffer);
            sourceFrame.LastBodyFrame.CopyData(_BodyBuffer);
        }
    }

    public class KinectAccessManager : IDisposable
    {
        private KinectSensor _Sensor;
        private CoordinateMapper _CoordinateMapper;
        private MultiSourceFrameReader _SourceFrameReader;

        private bool _IsSensorOpen = false;
        private bool _IsFrameReaderOpen = false;
        private bool _NullFrame = false;

        public bool IsAccess
        { get { return _IsSensorOpen && _IsFrameReaderOpen; } }
        public CoordinateMapper CoordinateMapper
        { get { return IsAccess ? _CoordinateMapper : null; } }
        public MultiSourceFrame SourceFrame
        {
            get
            {
                if (IsAccess)
                {
                    var val = _SourceFrameReader.AcquireLatestFrame();
                    _NullFrame = (val == null);
                    return val;
                }
                _NullFrame = true;
                return null;
            }
        }
        public bool NullFrame
        { get { return _NullFrame; } }


        public void Access()
        {
            OpenSensor();
            OpenFrameReader();

            _CoordinateMapper = _Sensor.CoordinateMapper;
        }

        private void OpenSensor()
        {
            if (_IsSensorOpen)
                return;

            _Sensor = KinectSensor.GetDefault();
            if (_Sensor == null)
            {
                //TODO: Popup Modal Window
                Debug.LogError("No ready Kinect found!");
                return;
            }

            _Sensor.Open();
            if (!_Sensor.IsOpen)
            {
                //TODO: Popup Modal Window
                Debug.LogError("Kinect sensor has problem!");
                CloseSensor();
                return;
            }

            _IsSensorOpen = true;
        }

        private void OpenFrameReader()
        {
            if (_IsFrameReaderOpen)
                return;

            _SourceFrameReader = _Sensor.OpenMultiSourceFrameReader(
                FrameSourceTypes.Color |
                FrameSourceTypes.Depth |
                FrameSourceTypes.BodyIndex |
                FrameSourceTypes.Body);

            _IsSensorOpen = (_SourceFrameReader != null);
        }

        public void Dispose()
        {
            CloseFrameReader();
            CloseSensor();
        }

        private void CloseSensor()
        {
            if (_IsSensorOpen)
            {
                _Sensor.Close();
                _IsSensorOpen = false;
            }
        }

        private void CloseFrameReader()
        {
            if (_IsFrameReaderOpen)
            {
                _SourceFrameReader.Dispose();
                _IsFrameReaderOpen = false;
            }
        }
    }

    public interface IMultiSourceFrame
    {
        IColorFrame LastColorFrame
        { get; }
        IBodyIndexFrame LastBodyIndexFrame
        { get; }
        IDepthFrame LastDepthFrame
        { get; }
        IBodyFrame LastBodyFrame
        { get; }
    }

    public interface IColorFrame
    {
        void CopyData(byte[] buffer);
    }

    public interface IBodyIndexFrame
    {
        void CopyData(byte[] buffer);
    }

    public interface IDepthFrame
    {
        void CopyData(ushort[] buffer);
    }

    public interface IBodyFrame
    {
        void CopyData(Body[] buffer);
    }

    public class RealMultiSourceFrame : IMultiSourceFrame
    {
        MultiSourceFrame _SourceFrame;
        RealColorFrame _ColorFrame;
        RealBodyIndexFrame _BodyIndexFrame;
        RealDepthFrame _DepthFrame;
        RealBodyFrame _BodyFrame;

        public IColorFrame LastColorFrame
        { get { return _ColorFrame; } }
        public IBodyIndexFrame LastBodyIndexFrame
        { get { return _BodyIndexFrame; } }
        public IDepthFrame LastDepthFrame
        { get { return _DepthFrame; } }
        public IBodyFrame LastBodyFrame
        { get { return _BodyFrame; } }

        public RealMultiSourceFrame(MultiSourceFrame sourceFrame)
        {
            _SourceFrame = sourceFrame;
            _ColorFrame = new RealColorFrame(sourceFrame.ColorFrameReference.AcquireFrame());
            _BodyIndexFrame = new RealBodyIndexFrame(sourceFrame.BodyIndexFrameReference.AcquireFrame());
            _DepthFrame = new RealDepthFrame(sourceFrame.DepthFrameReference.AcquireFrame());
            _BodyFrame = new RealBodyFrame(sourceFrame.BodyFrameReference.AcquireFrame());
        }

        private class RealColorFrame : IColorFrame
        {
            private ColorFrame _Frame;

            public RealColorFrame(ColorFrame frame)
            {
                _Frame = frame;
            }

            public void CopyData(byte[] buffer)
            {
                if (_Frame != null)
                {
                    var ColorData = GCHandle.Alloc(buffer, GCHandleType.Pinned);
                    _Frame.CopyConvertedFrameDataToIntPtr(ColorData.AddrOfPinnedObject(), (uint)buffer.Length, ColorImageFormat.Rgba);
                    ColorData.Free();
                }
            }
        }

        private class RealBodyIndexFrame : IBodyIndexFrame
        {
            private BodyIndexFrame _Frame;

            public RealBodyIndexFrame(BodyIndexFrame frame)
            {
                _Frame = frame;
            }

            public void CopyData(byte[] buffer)
            {
                if (_Frame != null)
                {
                    var BodyIndexData = GCHandle.Alloc(buffer, GCHandleType.Pinned);
                    _Frame.CopyFrameDataToIntPtr(BodyIndexData.AddrOfPinnedObject(), (uint)buffer.Length);
                    BodyIndexData.Free();
                }
            }
        }

        private class RealDepthFrame : IDepthFrame
        {
            private DepthFrame _Frame;

            public RealDepthFrame(DepthFrame frame)
            {
                _Frame = frame;
            }

            public void CopyData(ushort[] buffer)
            {
                if (_Frame != null)
                {
                    var DepthData = GCHandle.Alloc(buffer, GCHandleType.Pinned);
                    _Frame.CopyFrameDataToIntPtr(DepthData.AddrOfPinnedObject(), (uint)buffer.Length* sizeof(ushort));
                    DepthData.Free();
                }
            }
        }

        private class RealBodyFrame : IBodyFrame
        {
            private BodyFrame _Frame;

            public RealBodyFrame(BodyFrame frame)
            {
                _Frame = frame;
            }

            public void CopyData(Body[] buffer)
            {
                if (_Frame != null)
                    _Frame.GetAndRefreshBodyData(buffer);
            }
        }
    }
}
